/* ==========================================================================
 * delegate.c - Polymorphic libevent callbacks.
 * --------------------------------------------------------------------------
 * Copyright (c) 2009  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <errno.h>	/* EINVAL errno  */

#include <sys/time.h>	/* struct timeval */

#include <event.h>	/* EV_READ EV_WRITE event_set(3) event_add(3) event_del(3) event_pending(3)  */

#include "delegate.h"


#ifndef lengthof
#define lengthof(a)	(sizeof (a) / sizeof (a)[0])
#endif


void event_delegate_del(struct event_delegate *del) {
	if (del->uninstall) {
		*del->uninstall = 0;
		del->uninstall  = 0;
	}

	if (del->reinstall) {
		*del->reinstall = 0;
		del->reinstall  = 0;
	}

	if (event_pending(&del->event, EV_READ|EV_WRITE, 0))
		event_del(&del->event);

	del->fd      = -1;
	del->events  = 0;

	if (evtimer_pending(&del->timer, 0))
		event_del(&del->timer);

	del->timeout = (struct timeval){ 0 };
} /* event_delegate_del() */


static void event_delegate_fwd(int fd, short events, void *arg) {
	struct event_delegate *del = arg;
	int uninstall = !(del->events & EV_PERSIST);
	int reinstall = ((EV_PERSIST|EV_TIMEOUT) == (del->events & (EV_PERSIST|EV_TIMEOUT)));

	if (del->uninstall)
		*del->uninstall = 0;

	del->uninstall = &uninstall;

	if (del->cb)
		del->cb(fd, events, del->arg);
	else
		invoke(&del->del);

	if (uninstall)
		event_delegate_del(del);

	if (reinstall && !evtimer_pending(&del->timer, 0)) {
		evtimer_set(&del->timer, &event_delegate_fwd, del);
		evtimer_add(&del->timer, &del->timeout);
	}
} /* event_delegate_fwd() */


int event_delegate_add(struct event_delegate *del, int fd, short events, struct timeval *timeout) {
	int error;

	if (del->uninstall) {
		*del->uninstall = 0;
		del->uninstall  = 0;
	}

	if (del->fd != fd || del->events != events)
		event_delegate_del(del);

	if (evtimer_pending(&del->timer, 0)) {
		event_del(&del->timer);

		del->timeout = (struct timeval){ 0 };
	}

	if (fd != -1 && events && !event_pending(&del->event, EV_READ|EV_WRITE, 0)) {
		event_set(&del->event, fd, events|EV_PERSIST, &event_delegate_fwd, del);

		if (0 != event_add(&del->event, 0))
			goto syerr;

		del->fd     = fd;
		del->events = events;
	}

	if (timeout) {
		evtimer_set(&del->timer, &event_delegate_fwd, del);
		evtimer_add(&del->timer, timeout);

		del->timeout = *timeout;
		del->events  = events;
	}

	return 0;
syerr:
	error = errno;
error:
	event_delegate_del(del);

	return error;
} /* event_delegate_add() */


_Bool event_delegate_pending(struct event_delegate *del, short events, struct timeval *timeout) {
	if (events & (EV_READ|EV_WRITE)) {
		if (event_pending(&del->event, events & (EV_READ|EV_WRITE), timeout))
			return 1;
	}

	if (events & EV_TIMEOUT) {
		if (event_pending(&del->timer, events & EV_TIMEOUT, timeout))
			return 1;
	}

	return 0;
} /* event_delegate_pending() */


#if DELEGATE_MAIN

#include <stdio.h>


static void func(long long ll, const char *str, struct delegate *del) {
	static int depth;

	depth++;

	printf("ll:%lld str:[%s] del:%p depth:%d\n", ll, str, (void *)del, depth);

	if (depth < 3)
		invoke(del);

	depth--;
} /* func() */


int main(int argc, char *argv[]) {
	char *hello = "hello world again";
	struct delegate del;

	invoke(delegate(&del, &func, 0xdeafbeefLL, "hello world", &del));

	invoke(delegate(&del, &func, 0x0ffffffffffffff0LL, hello, &del));

	return 0;
} /* main() */

#endif /* DELEGATE_MAIN */
